#ifndef _BASICBC_H
#define _BASICBC_H

#include "DistributionLibrairie.h"
#include "BasicBCIdl.h"

#include <list>

using namespace std;

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

typedef enum { BASICBC_BLOC=1, // no parameter: all bloc same size but last one: bc(n+p-1/p)
	       BASICBC_CYCLIC,  // no parameter: bc(1)
	       BASICBC_BLOCCYCLIC // parameter
} BasicBC_distri_type_t;

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

typedef struct {
  
  BasicBC_distri_type_t type;
  unsigned long unitsize;  // size of one element of our "vector"; in octet
  unsigned long blocsize; // for bloc-cyclic; in element

} BasicBC_param_t;

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

class BasicBC : public DistributionLibrary
{
private:
  ////////////////////////////////////////
  // Client & Server side

  unsigned _config; // vector of bit describing which elements have been configured (cf .cc)

  PaCO::PacoTopology_t _sTopo; // topology of the source (sender)  -- SET
  PaCO::PacoTopology_t _dTopo; // topology of the destination (receiver) -- SET

  BasicBC_param_t _param;    // the parameter of the distribution -- SET
  unsigned _nodeRank;        // id of the node (either source or destination) -- SET

  unsigned _glen;     // total vector size: in element -- SET
  unsigned _llen;     // local vector size: in element -- COMPUTED

  unsigned _lstart;   // id of 1st local element    -- COMPUTED
  unsigned _sbsz;     // source blocsize in element -- COMPUTED
  unsigned _dbsz;     // source blocsize in element -- COMPUTED
  
  ////////////////////////////////////////
  // Client Side

  // Info passed between computePiecesToSend & getClientData
  // is in fact a map of info_t with destNode as key :(
  // Private type
  typedef struct {
    unsigned gstart;    // in element
    unsigned lstart;    // in octet
    unsigned msg_size;  // in octet
    unsigned sent_size; // in octet 
  } info_t;
  typedef list<info_t*> info_list_t;

  typedef map<long, info_list_t**> info_list_map_t; // also store com for id == -1
  info_list_map_t _info_list_map;

  long _cur_id; // current communication id, -1 means none
  info_list_t** _infolists; // current array of dTotal entries

  bool _descr_to_be_cleaned; // true when descr should be reset - cf getClientData & getDescr (during an invocation)
  char* _clientBuffer; // pointer to the user data (set)
  BasicBCLib::BasicBCDescr _clientDescr; // contains the descr exchange between sender & destination -- SET indirectly

  ////////////////////////////////////////
  // Server Side
  unsigned long _serverToReceived; // #of element yet to be received (during an invocation)
  unsigned _descr_ids_counter; // id of the current block (msg = sequence<block>) (during an invocation)
  char* _serverBuffer; // Buffer where to receive the data

  BasicBCLib::BasicBCDescr* _serverDescr; // contains the descr exchange between sender & destination -- computed

  ////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////

public:
  BasicBC();
  virtual ~BasicBC();

  // Common
  virtual void setSourceTopology(PaCO::PacoTopology_t topo); 
  virtual PaCO::PacoTopology_t getSourceTopology(); 
  virtual void setDestTopology(PaCO::PacoTopology_t topo); 
  virtual PaCO::PacoTopology_t getDestTopology(); 
  virtual void setNodeRank(long Rank);
  virtual long getNodeRank(); 
  virtual void setCommunicator(void * group) {}

  // Data description API
  void setEltSize(unsigned long size);
  void setBlocSize(unsigned long bsz); // bloc-> bc(0), cyclic -> bc(1)
  void setTotalNbElt(unsigned long elt_nb);

  // Client Side
  virtual PieceToSend* computePiecesToSend(unsigned& size);
  virtual void  setDataPtr(void* dataPtr);
  virtual void* getClientData(void *pid, int server_node, long & size, long & length, bool & end);
  virtual void* getDescr();
  virtual void  clientFree(); // true if com schedule to be freed

  virtual bool setComId(long id);
  virtual bool freeComId(long id);

  // Server Side
  virtual bool  insertData(void* rcptBuf, unsigned long element_nb);
  virtual void  setDescr(void * descr);

  virtual void* getServerData(long & length);
  virtual void  serverFree();

private:  
  void internalSetComId(long id);
  void internalFreeComId(info_list_map_t::iterator& it);

  // Helper function
  void allocate_new_info_list();
  void reset_info_list();
  void clear_info_list();

};

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

class BasicBC_fab : virtual public paco_fabrique_distribution
{

 public:
  BasicBC_fab() {}
  virtual ~BasicBC_fab() {}
  
  DistributionLibrary *create()
    {
      return new BasicBC();
    }

};

#endif
